from pathlib import Path
import pytest
from gentest.exec import validate_query_results, path_to_slug

basic_dir = Path(__file__).parent / "basic"
basic_files = [path for path in basic_dir.iterdir() if path.suffix == ".py"]

@pytest.mark.parametrize("file_path", basic_files, ids=lambda path: path_to_slug(path, basic_dir))
def test_basic(file_path: Path, snapshot):
    validate_query_results(file_path, snapshot)



from pathlib import Path
import pytest
from gentest.util import PROJECT_DIR
from gentest.exec import validate_query_results, path_to_slug

example_dir = Path(PROJECT_DIR) / "examples"
example_files = [path for path in example_dir.iterdir() if path.suffix == ".py"]

@pytest.mark.parametrize("file_path", example_files, ids=lambda path: path_to_slug(path, example_dir))
def test_example(file_path: Path, snapshot):
    validate_query_results(file_path, snapshot)



from collections import defaultdict
import json
from typing import Any, Dict, Iterable, List, Set, cast
from datetime import datetime, date
import pandas as pd
from zoneinfo import ZoneInfo

from .metamodel import Behavior, Builtins, Action, ActionType, Namer, Property, Type, Var, Task, Value
from . import metamodel as m
from . import compiler as c
from .clients import config
from .dsl import build
import re

gather_vars = m.Utils.gather_vars

#--------------------------------------------------
# OrderedSet
#--------------------------------------------------

class OrderedSet:
    def __init__(self):
        self.set:Set[Var] = set()
        self.list:List[Var] = []

    def add(self, item):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)

    def update(self, items:Iterable[Any]):
        for item in items:
            self.add(item)

    def __contains__(self, item):
        return item in self.set

    def __bool__(self):
        return bool(self.set)

    def __iter__(self):
        return iter(self.list)

#--------------------------------------------------
# Flow
#--------------------------------------------------

class Flow():
    def __init__(self):
        self.tasks:List[Task] = []
        self.inlines:Set[Task] = set()
        self.task_bindings:Dict[Task, OrderedSet] = defaultdict(OrderedSet)
        self.task_deps:Dict[Task, Set[Task]] = defaultdict(set)
        self.provider_stack:List[Dict[Var, Task]] = [{}] # we start with a global scope
        self.task_var_mapping:Dict[Task, Dict[Var, Var]] = defaultdict(dict)

    def push_context(self):
        self.provider_stack.append({})
        neue = Task()
        self.tasks.append(neue)
        return neue

    def pop_context(self, task:Task, inline=False, ignore_deps=False, mappings:Dict[Var, Var]=dict()):
        if not ignore_deps:
            vs = [mappings.get(v, v) for v in gather_vars(task.items)]
            existing = self.fetch(vs, task)
            for e in existing:
                if e != task:
                    self.task_deps[task].add(e)
        if inline:
            self.inlines.add(task)
        self.provider_stack.pop()

    def assoc(self, vars:Iterable[Var], task:Task):
        for var in vars:
            self.provider_stack[-1][var] = task

    def has_var(self, var:Var):
        for provider in reversed(self.provider_stack):
            if var in provider:
                return True
        return False

    def fetch(self, vars:Iterable[Var], cur:Task):
        tasks:Set[Task] = set()
        for var in vars:
            for provider in reversed(self.provider_stack):
                found = provider.get(var)
                if found and found != cur:
                    tasks.add(found)
                    self.task_bindings[found].add(var)
                    break
        return tasks

    def mapped_bindings(self, task:Task, dep:Task, vars:Iterable[Var]|None=None):
        mapped = self.task_var_mapping.get(task)
        bindings = vars or self.task_bindings[dep]
        if mapped and len(mapped):
            return [mapped.get(b, b) for b in bindings]
        return bindings

    def finalize(self):
        final_tasks = []
        for neue in self.tasks:
            for dep in self.task_deps[neue]:
                if dep in self.tasks:
                    neue.items.insert(0, build.relation_action(ActionType.Get, dep, self.mapped_bindings(neue, dep)))
            if self.task_bindings[neue]:
                neue.items.append(build.relation_action(ActionType.Bind, neue, self.task_bindings[neue]))
            if neue not in self.inlines:
                final_tasks.append(neue)
        return final_tasks

    def reset(self):
        self.tasks.clear()
        self.provider_stack.clear()
        self.provider_stack.append({})
        self.task_bindings.clear()

#--------------------------------------------------
# Dataflow
#--------------------------------------------------

class Dataflow(c.Pass):
    def __init__(self, copying=True) -> None:
        super().__init__(copying)
        self.flow = Flow()

    def reset(self):
        super().reset()
        self.flow.reset()

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, task:Task, parent):
        self.query_flow(task)
        if not parent:
            final_tasks = self.flow.finalize()
            task.behavior = Behavior.Sequence
            task.items = [build.relation_action(ActionType.Call, task, []) for task in final_tasks]

    def query_flow(self, task:Task, inline=False, ignore_deps=False):
        flow = self.flow
        neue:Task = flow.push_context()

        for orig in task.items:
            ent:Any = orig.entity
            if orig.action == ActionType.Call and self.local_task(ent):
                flow.assoc(gather_vars(neue.items), neue)
                behavior = ent.value.behavior
                if behavior == Behavior.Union:
                    neue = self.union_call(orig, neue)
                elif behavior == Behavior.OrderedChoice:
                    neue = self.ordered_choice_call(orig, neue)
                elif behavior == Behavior.Query:
                    self.walk(orig, task)
            elif ent.isa(Builtins.Quantifier):
                neue = self.quantifier_call(orig, neue)
            elif ent.isa(Builtins.Aggregate):
                neue = self.aggregate_call(orig, neue)
            else:
                neue.items.append(orig)

        flow.pop_context(neue, inline=inline, ignore_deps=ignore_deps)
        return neue

    def local_task(self, var:Var):
        return isinstance(var.value, Task) and len(var.value.items) > 0

    #--------------------------------------------------
    # Union
    #--------------------------------------------------

    def union_call(self, call:Action, parent:Task):
        has_dep = False
        # run through the subtasks and add them to the flow
        for item in cast(Task, call.entity.value).items:
            neue = self.query_flow(cast(Task, item.entity.value))
            has_dep = has_dep or (parent in self.flow.task_deps[neue])
            rets = [i for i in neue.items if i.entity.value == Builtins.Return]
            if len(rets):
                rets[0].entity.value = call.entity.value
                rets[0].action = ActionType.Bind

        # if one of the subtasks depends on the parent, we need to cut the
        # current task to prevent cycles and carry on in a new one
        if has_dep:
            orig = parent
            self.flow.pop_context(orig)
            self.flow.assoc(gather_vars(orig.items), orig)
            parent = self.flow.push_context()
        parent.items.append(build.relation_action(ActionType.Get, cast(Task, call.entity.value), call.bindings.values()))
        return parent

    #--------------------------------------------------
    # Ordered choice
    #--------------------------------------------------

    def ordered_choice_call(self, call:Action, parent:Task):
        ordered_choice = cast(Task, call.entity.value)
        has_dep = False
        prevs = []
        branches = []

        # depending on whether or not the the subtasks depend on the parent,
        # we'll need to update all the references to the output so that any vars
        # we depend on are included in the result, to guarantee that the choosen
        # values join with the correct original rows
        result_refs = [build.relation_action(ActionType.Get, ordered_choice, call.bindings.values())]
        for item in ordered_choice.items:
            neue:Task = self.query_flow(cast(Task, item.entity.value))
            branches.append(neue)
            # find the return statement, we'll turn it into a bind for the overall orderd_choice
            rets = [i for i in neue.items if i.entity.value == Builtins.Return]
            ret_bindings = []
            if len(rets):
                rets[0].entity.value = neue
                result_refs.append(rets[0])
                ret_bindings = rets[0].bindings.values()
            else:
                rets.append(build.relation_action(ActionType.Bind, neue, ret_bindings))
                neue.items.append(rets[0])
                result_refs.append(rets[0])

            # add a bind for this particular task so that we can negate it in subsequent
            # branches, allowing us to create the ordering
            bind = build.relation_action(ActionType.Bind, ordered_choice, ret_bindings)
            result_refs.append(bind)
            neue.items.append(bind)
            # clear the ret bindings, since this should just return whether or not this
            # branch was successful
            rets[0].bindings.clear()
            # Negate all the previous branches to ensure we only return a value if
            # we're at our position in the order
            for prev in prevs:
                fetch = build.relation_action(ActionType.Get, prev, [])
                result_refs.append(fetch)
                prev_task = Task(items=[fetch])
                neue.items.append(build.call(Builtins.Not, [Var(value=[]), Var(value=prev_task)]))
            prevs.append(neue)

            has_dep = has_dep or (parent in self.flow.task_deps[neue])

        # If one of the branches depends on the parent task, we need to cut the parent
        # and create a new one to prevent cycles
        if has_dep:
            orig = parent
            self.flow.pop_context(orig)
            self.flow.assoc(gather_vars(orig.items), orig)
            # We also need to make sure that branches that don't currently depend on
            # the parent now do, since they also need to join correctly with the original
            # rows
            for branch in branches:
                self.flow.task_deps[branch].add(orig)
            # Update the bindings for the result refs to include the vars we depend on
            for bind in result_refs:
                prop_len = len(bind.bindings)
                for i, var in enumerate(self.flow.task_bindings[orig]):
                    bind.bindings[Builtins.Relation.properties[prop_len+i]] = var
            parent = self.flow.push_context()

        parent.items.append(result_refs[0]) # result_refs[0] is the Get for the ordered_choice
        return parent

    #--------------------------------------------------
    # Quantifiers
    #--------------------------------------------------

    def quantifier_call(self, call:Action, parent:Task):
        quantifier = cast(Task, call.entity.value)
        group, task_var = [*call.bindings.values()]
        sub_task = self.query_flow(cast(Task, task_var.value))

        if isinstance(group.value, list) and len(group.value):
            raise Exception("TODO: grouped quantifiers")

        # Find any vars that this quantified task depends on
        sub_vars = gather_vars(sub_task.items)
        parent_vars = gather_vars(parent.items)
        # Find vars that we explicitly fetch in the child task that are used in the same
        # ent.attr pair in the parent task, so we can remove them as dependencies
        # this allows person.friend.name == "Joe" to not unify with an outer person.friend
        # constraint
        for item in sub_task.items:
            if item.action == ActionType.Get and item.entity in parent_vars:
                sub_vars -= set(item.bindings.values())
        shared = sub_vars & parent_vars
        shared.update([var for var in sub_vars if self.flow.has_var(var)])
        # bind those so we can use them in the quantified task
        sub_task.items.append(build.relation_action(ActionType.Bind, sub_task, shared))

        # create the quantified task, which just gets the subtask
        quantifed_task = Task()
        quantifed_task.items.append(build.relation_action(ActionType.Get, sub_task, shared))

        # add the call to the quantifier
        parent.items.append(build.call(quantifier, [group, Var(value=quantifed_task)]))
        return parent

    #--------------------------------------------------
    # Aggregates
    #--------------------------------------------------

    def aggregate_call(self, call:Action, parent:Task):
        orig = parent
        self.flow.pop_context(orig)
        self.flow.assoc(gather_vars(orig.items), orig)
        agg = cast(Task, call.entity.value)
        (args, group, ret) = [*call.bindings.values()]
        group_vars = cast(List[Var], group.value)

        # create the inner relation we'll aggregate over
        inner = self.flow.push_context()
        inner_vars = cast(List[Var], args.value)
        # to prevent shadowing errors we need to map the inner vars to new vars
        mapped = [Var(name=var.name, type=var.type) for var in inner_vars]
        self.flow.task_var_mapping[inner] = dict(zip(inner_vars, mapped))
        # vars that are in both the projection and grouping needed to be mapped in
        # the projection but made equivalent in the body so the grouping takes effect
        equivs = [(orig, neue) for (orig, neue) in self.flow.task_var_mapping[inner].items() if orig in group_vars]
        for (orig, neue) in equivs:
            inner.items.append(build.relation_action(ActionType.Call, Builtins.eq, [orig, neue]))
        # bind the mapped vars as the output of the inner relation
        inner.items.append(build.relation_action(ActionType.Bind, inner, mapped))
        self.flow.pop_context(inner, inline=True, mappings=dict(zip(mapped, inner_vars)))

        # create the outer aggregate
        outer = self.flow.push_context()
        outer_call = build.relation_action(ActionType.Call, agg, [Var(value=inner), ret])
        outer.items.append(outer_call)
        if agg.isa(Builtins.Extender):
            self.flow.task_bindings[outer].add(ret)
            for g in group_vars:
                self.flow.task_bindings[outer].add(g)
            for ix, var in enumerate(inner_vars):
                self.flow.task_bindings[outer].add(var)
                outer_call.bindings[Builtins.Relation.properties[ix+2]] = var
        else:
            for g in group_vars:
                self.flow.task_bindings[outer].add(g)
            self.flow.task_bindings[outer].add(ret)
        self.flow.pop_context(outer, ignore_deps=True)

        # Resume the flow
        parent = self.flow.push_context()
        parent.items.append(build.relation_action(ActionType.Get, outer, self.flow.task_bindings[outer]))
        return parent


#--------------------------------------------------
# Shredder
#--------------------------------------------------

class Shredder(c.Pass):
    def query(self, task: Task, parent):
        neue_actions = []
        for item in task.items:
            if item.action not in [ActionType.Call, ActionType.Construct] and not item.entity.isa(Builtins.Relation):
                ident, action = item.entity, item.action
                for type in item.types:
                    neue_actions.append(build.relation_action(action, type, [ident]))
                for prop, value in item.bindings.items():
                    neue_actions.append(build.relation_action(action, prop, [ident, value]))
            else:
                neue_actions.append(item)
        task.items = neue_actions

#--------------------------------------------------
# Splinter
#--------------------------------------------------

class Splinter(c.Pass):

    def query(self, task: Task, parent):
        effects = [i for i in task.items if i.action.is_effect()]

        grouped_effects = defaultdict(list)
        for item in effects:
            grouped_effects[(item.action, item.entity.value)].append(item)

        if len(grouped_effects) > 1:
            neue_items = []

            non_effects = [i for i in task.items if not i.action.is_effect()]
            effects_vars = gather_vars(effects)

            fetch = None
            if len(non_effects):
                fetch = self.create_fetch(non_effects, effects_vars)
                neue_items.append(fetch)

            for (k, b) in grouped_effects.items():
                neue_items.append(self.create_effect_query(b, effects_vars, fetch.entity.value if fetch else None))

            task.behavior = Behavior.Sequence
            task.items = neue_items

    #--------------------------------------------------
    # Subtask creation
    #--------------------------------------------------

    def create_fetch(self, non_effects: List[Action], effects_vars: Iterable[Var]):
        fetch = Task()
        non_effects.append(build.relation_action(ActionType.Bind, fetch, effects_vars))
        fetch.items = non_effects
        return build.call(fetch, [])

    def create_effect_query(self, effects: List[Action], effects_vars: Iterable[Var], fetch: Any):
        neue = Task()
        if fetch:
            effects.insert(0, build.relation_action(ActionType.Get, fetch, effects_vars))
        neue.items = effects
        return build.call(neue, [])

#--------------------------------------------------
# SetCollector
#--------------------------------------------------

set_types = [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]

class SetCollector(c.Pass):
    def query(self, query: Task, parent):
        binds = [i for i in query.items if i.action in set_types]
        if len(binds) > 1:
            neue_items = []
            for item in query.items:
                if item.action not in set_types:
                    neue_items.append(item)
            neue_items.extend(self.create_raw(binds))
            query.items = neue_items

    def create_raw(self, binds: List[Action]):
        vals = [Var(value=[]) for i in range(len(binds[0].bindings))]
        vars = [Var() for v in vals]

        for bind in binds:
            for ix, var in enumerate(bind.bindings.values()):
                cast(List[Var], vals[ix].value).append(var)

        return [
            build.relation_action(ActionType.Get, Builtins.RawData, vals + vars),
            build.relation_action(binds[0].action, cast(Type, binds[0].entity.value), vars)
        ]

#--------------------------------------------------
# Emitter
#--------------------------------------------------

rel_keyword = ["and", "or", "not", "def", "end"]
rel_infix = [">", "<", ">=", "<=", "=", "!=", "+", "-", "*", "/", "^"]

def sanitize(input_string, is_rel_name = False):
    # Replace non-alphanumeric characters with underscores
    if is_rel_name and input_string.endswith("]"):
        sanitized = re.sub(r'[^\w:\[\]" ,]|^(?=\d)', "_", input_string)
    else:
        sanitized = re.sub(r"[^:\w]|^(?=\d)", "_", input_string)

    # Check if the resulting string is a keyword and append an underscore if it is
    if sanitized in rel_keyword:
        sanitized += "_"

    return sanitized

class Emitter(c.Emitter):

    def __init__(self):
        super().__init__()
        self.namer = Namer()
        self.stack:List[Set[Var]] = []
        self.mapped = {}

    #--------------------------------------------------
    # Emit
    #--------------------------------------------------

    def emit(self, task: Task|Var):
        self.mapped.clear()
        code = ""
        try:
            if isinstance(task, Task):
                code = getattr(self, task.behavior.value)(task)
            elif isinstance(task, Var):
                code = self.emit_var(task)
        except Exception as e:
            print("EMIT FAILED:", e)
            # raise e
        return code

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def sanitize(self, input_string, is_rel_name = False):
        return sanitize(input_string, is_rel_name)

    #--------------------------------------------------
    # Vars
    #--------------------------------------------------

    def emit_val(self, value: Value, var:Var|None=None):
        if value is True:
            return '"True"'
        if isinstance(value, list):
            return f"{', '.join([self.emit_var(v) for v in value])}"
        if isinstance(value, Task):
            return self.to_inline_relation(value)
        if isinstance(value, bytes):
            if var and var.type.isa(Builtins.ValueType):
                return f"^{self.sanitize(var.name).capitalize()}Type[0x{value.hex()}]"
            return f"0x{value.hex()}"
        if isinstance(value, Property):
            return self.sanitize(value.name)
        if isinstance(value, Type):
            return self.sanitize(value.name, value.isa(Builtins.Relation))
        if isinstance(value, pd.Timestamp):
            t = value.tz_localize('UTC')
            return t.isoformat()
        if isinstance(value, datetime):
            t = value.astimezone(ZoneInfo('UTC'))
            return t.isoformat()
        if isinstance(value, date):
            t = value
            return t.isoformat()
        if isinstance(value, str):
            # % is the string interpolation character in rel
            return json.dumps(value).replace("%", "\\%")
        return json.dumps(value)

    def emit_var(self, var: Var|Value):
        if not isinstance(var, Var):
            return self.emit_val(var)
        if var in self.mapped:
            return self.mapped[var]
        if var.type.isa(Builtins.Symbol):
            return f":{self.sanitize(var.value)}"
        if var.value is not None:
            return self.emit_val(var.value, var)
        if var.name == "_":
            return "_"
        name = self.namer.get(var)
        return f"_{self.sanitize(name).lower()}"

    #--------------------------------------------------
    # Sequence
    #--------------------------------------------------

    def sequence_action(self, action: Action):
        if action.entity.value == Builtins.RawCode:
            return str([*action.bindings.values()][0].value)
        elif isinstance(action.entity.value, Task):
            return self.emit(action.entity.value)
        else:
            raise Exception(f"TODO: Rel emit for action type {action.action}")

    def sequence(self, task: Task):
        items = [ self.sequence_action(i) for i in task.items ]
        return "\n\n".join(items)

    #--------------------------------------------------
    # Query helpers
    #--------------------------------------------------

    def to_relation(self, action: Action, vars: set, body:List[str], in_head = False):
        root = cast(Type|Property, action.entity.value)
        rel_name = root.name
        if not rel_name and isinstance(root, Task):
            rel_name = f"T{root.id}"
        args = []
        for var in action.bindings.values():
            emitted = self.emit_var(var)
            vars.add(var)
            if in_head and emitted in args:
                name = self.namer.get(var)
                orig = emitted
                emitted = f"_{self.namer.get_safe_name(name)}"
                body.append(f"{emitted} = {orig}")
            elif in_head and var.value is not None and var.isa(Builtins.ValueType):
                name = self.namer.get(var)
                orig = emitted
                emitted = f"_{self.namer.get_safe_name(name)}"
                body.append(f"{emitted} = {self.emit_val(var.value, var)}")
            args.append(emitted)
        if rel_name in rel_infix:
            if len(args) == 2:
                return f"{args[0]} {rel_name} {args[1]}"
            else:
                return f"{args[2]} = {args[0]} {rel_name} {args[1]}"
        rel_name = self.sanitize(rel_name, True)
        if isinstance(root, Task) and root.isa(Builtins.Quantifier):
            # TODO: handle quantifier grouping
            args = args[1:]
            rel_name = rel_name.lower()
        if action.action == ActionType.Persist:
            rel_name = f"insert:{rel_name}"
        elif action.action == ActionType.Unpersist:
            rel_name = f"delete:{rel_name}"
        final = f"{rel_name}({', '.join(args)})"
        return final

    def to_return(self, action: Action, vars: set, body:List[str]):
        args = []
        for var in action.bindings.values():
            emitted = self.emit_var(var)
            vars.add(var)
            if emitted in args:
                name = self.namer.get(var)
                orig = emitted
                emitted = f"_{self.namer.get_safe_name(name)}"
                body.append(f"{emitted} = {orig}")
            args.append(emitted)
        return f"output({', '.join(args)})"

    def to_set(self, action: Action, vars: set):
        vals:Any = [v for v in action.bindings.values()]
        mid_point = len(vals) // 2
        vars_str = ", ".join([self.emit_var(i) for i in vals[mid_point:]])
        rows = []
        for ix in range(len(vals[0].value)):
            rows.append(",".join([self.emit_var(col.value[ix]) for col in vals[:mid_point]]))
        row_str = "; ".join(rows)
        return f"{{{row_str}}}({vars_str})"

    def to_inline_set(self, action: Action, vars: set):
        vals:Any = [v for v in action.bindings.values()]
        rows = []
        for ix in range(len(vals[0].value)):
            rows.append(",".join([self.emit_var(col.value[ix]) for col in vals[:-1]]))
        row_str = "; ".join(rows)
        return (vals[-1], f"{{{row_str}}}")

    def to_inline_relation(self, task:Task, existing_vars=set()):
        return f"( {self.query(task, True)} )"

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, task: Task, inline = False):
        supporting_rules = []
        head = ""
        body = []
        body_vars = set()
        head_vars = set()

        task_vars = gather_vars(task.items)
        self.stack.append(task_vars)

        for i in task.items:
            if i.action in [ActionType.Get, ActionType.Call]:
                if i.entity.value == Builtins.RawData:
                    body.append(self.to_set(i, body_vars))
                elif i.entity.value == Builtins.InlineRawData:
                    (v, data) = self.to_inline_set(i, body_vars)
                    self.mapped[v] = data
                    head_vars.add(v)
                elif i.entity.value == Builtins.Install:
                    (item,) = i.bindings.values()
                    supporting_rules.append(f"bound {self.emit_var(item)}")
                    if isinstance(item.value, Type) and item.value.isa(Builtins.ValueType):
                        supporting_rules.append(f"value type {self.sanitize(item.value.name).capitalize()}Type = UInt128")
                else:
                    body.append(self.to_relation(i, body_vars, body))
            elif i.action == ActionType.Bind and i.entity.value == Builtins.Return:
                head_rel = self.to_return(i, head_vars, body)
                head = f"def {head_rel} =\n    "
            elif i.action in [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]:
                if not inline:
                    head_rel = self.to_relation(i, head_vars, body, True)
                    annotations = set()
                    ent = i.entity.value
                    if isinstance(ent, Type) and ent.isa(Builtins.Annotation):
                        for parent in ent.parents:
                            if parent.isa(Builtins.Annotation):
                                annotations.add(parent.name.lower())
                    if annotations:
                        annotation_str = " ".join([f"@{a}" for a in annotations])
                        head = f"{annotation_str}\ndef {head_rel} =\n    "
                    else:
                        head = f"def {head_rel} =\n    "
                else:
                    props = list(i.bindings.values())
                    head_vars.update(props)
                    head = ",".join([self.emit_var(v) for v in props]) + ": "
            elif i.action == ActionType.Construct:
                inputs = [*i.bindings.values()]
                input_str =', '.join([self.emit_var(v) for v in inputs[:-1]])
                output = self.emit_var(inputs[-1])
                body_vars.update(inputs)
                body.append(f"{output} = ^{self.sanitize(i.entity.value.name).capitalize()}Type[{input_str}]")
            else:
                raise Exception(f"TODO: Rel emit for action type {i.action}")
        body_str = " and\n    ".join(body)
        existing_vars = set()
        for vars in self.stack[:-1]:
            existing_vars.update(vars)
        from_vars = [self.emit_var(v) for v in (body_vars - head_vars - existing_vars) if v.value is None]
        if len(from_vars) and not inline:
            body_str += f"\n    from {', '.join(from_vars)}"
        if head and len(from_vars) and inline:
            body_str += f" from {', '.join(from_vars)}"


        self.stack.pop()

        if not head and inline:
            return body_str
        elif not head:
            head = f"def T{task.id}() =\n    "

        support_str = ""
        if len(supporting_rules):
            support_str = "\n\n".join(supporting_rules) + "\n\n"

        if support_str and not head_vars and not body_str:
            return support_str

        if not body_str:
            body_str = "{()}"


        return f"{support_str}{head}{body_str}"

#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Clone(c.Pass):
    pass

class Compiler(c.Compiler):
    def __init__(self, config:config.Config):
        super().__init__(Emitter(), [
            Clone(),
            Dataflow(),
            Shredder(),
            Splinter(),
            SetCollector(),
        ])



import os
from typing import Dict, Any, cast
from railib import config
import configparser
import toml
import tomlkit
from tomlkit.items import Table


#--------------------------------------------------
# config defaults
#--------------------------------------------------

FIELD_PLACEHOLDER = ""

snowflake_default_props = {
    "platform": "snowflake",
    "user": FIELD_PLACEHOLDER,
    "password": FIELD_PLACEHOLDER,
    "account": FIELD_PLACEHOLDER,
    "role": "PUBLIC",
    "warehouse": FIELD_PLACEHOLDER,
    "rai_app_name": FIELD_PLACEHOLDER,
    "engine": FIELD_PLACEHOLDER,
    "compute_pool": FIELD_PLACEHOLDER,
}

azure_default_props = {
    "platform": "azure",
    "host": "azure.relationalai.com",
    "port": 443,
    "region": "us-east",
    "scheme": "https",
    "client_credentials_url": "https://login.relationalai.com/oauth/token",
    "client_id": FIELD_PLACEHOLDER,
    "client_secret": FIELD_PLACEHOLDER,
}

all_prop_keys = set(snowflake_default_props.keys()) | set(azure_default_props.keys())

#--------------------------------------------------
# helpers
#--------------------------------------------------

class ConfigFile:
    def __init__(
        self,
        path: str | None = None,
        cfg: Dict[str, Any] = {},
        *,
        format = "toml"
    ):
        if format == "toml":
            self.path:str|None = path
            self.profiles:Dict[str,Dict[str,Any]] = cfg.get("profile", {})
            self.config:Dict[str,Any] = {}
            for key in cfg:
                if key != "profile":
                    self.config[key] = cfg[key]
        elif format == "ini":
            self.path = path
            self.profiles = {
                k: {k2: v2 for k2, v2 in v.items()}
                for k, v in cfg.items()
                if k != "DEFAULT"
            }
            self.config = {}
        else:
            raise ValueError(f"Unsupported format: {format}")

    def apply_keymap(self):
        keymap = {
            "active-profile": "active_profile",
            "snowflake-connection": "snowflake_connection",
            "snowsql_user": "user",
            "snowsql_pwd": "password",
        }
        new_config = {}
        for key in self.config:
            new_config[keymap.get(key, key)] = self.config[key]
        self.config = new_config
        new_profiles = {}
        for profile, props in self.profiles.copy().items():
            new_profiles[profile] = {}
            for key in props.copy():
                # self.profiles[profile][keymap.get(key, key)] = props[key]
                new_profiles[profile][keymap.get(key, key)] = props[key]
        self.profiles = new_profiles

    def get_combined_profiles(self):
        combined_profiles = {}
        combined_profiles["__top_level__"] = self.config or {}
        for profile in self.profiles:
            combined_profiles[profile] = self.profiles[profile].copy()
        return combined_profiles

    def map(self, f):
        """
        Apply a function `f` to all props dictionaries in the configuration files, including the top-level dictionary and all profiles.
        """
        combined_profiles = self.get_combined_profiles()
        for profile in combined_profiles:
            combined_profiles[profile] = f(combined_profiles[profile])
        config = combined_profiles.pop("__top_level__")
        return self.__class__(
            self.path, { **config, "profile": combined_profiles }
        )

    def filled_from_snowflake_connection(self):
        def fill(props):
            if "snowflake_connection" in props:
                connection_name = props.pop("snowflake_connection")
                snowflake_config = get_from_snowflake_connections_toml()
                if snowflake_config and connection_name in snowflake_config:
                    return {
                        **snowflake_config[connection_name],
                        **props,
                    }
            return props
        return self.map(fill)

    def merge(self, other):
        """
        Merge the profiles from `other` into this ConfigFile object
        """
        combined_profiles = self.get_combined_profiles()
        other_combined_profiles = other.get_combined_profiles()

        all_profile_names = set(combined_profiles) | set(other_combined_profiles)

        for profile in all_profile_names:
            self_profile = combined_profiles.get(profile, {})
            other_profile = other_combined_profiles.get(profile, {})
            combined_profiles[profile] = {**other_profile, **self_profile}

        config = combined_profiles.pop("__top_level__", None)

        if config is not None:
            self.config = config
        self.profiles = combined_profiles


def first(x):
    return next(iter(x), None)

def _search_upwards_for_file(file:str):
    """
    Search for `file` in the current directory and all parent directories.

    Returns the absolute path to the file if found, otherwise None.
    """
    dir = os.path.abspath(os.getcwd())
    while True:
        file_path = os.path.join(dir, file)
        if os.path.isfile(file_path):
            yield file_path
        parent_dir = os.path.dirname(dir)
        if parent_dir == dir:
            break # reached the root
        dir = parent_dir

def _search_userdir_for_file(file:str):
    """
    Search for `file` in the user's home directory.

    Returns the absolute path to the file if found, otherwise None.
    """
    file_path = os.path.expanduser(f"~/{file}")
    if os.path.isfile(file_path):
        yield file_path

def _find_config_file():
    yield from _search_upwards_for_file("raiconfig.toml")
    yield from _search_userdir_for_file(".rai/raiconfig.toml")

def _parse_and_map_config(file:str):
    """
    Parse the config file at `file` and return a dictionary of config values.

    Handles both TOML and INI files.
    """
    if not os.path.exists(file):
        return ConfigFile(cfg={})
    if file.endswith(".toml"):
        try:
            cf = ConfigFile(file, toml.load(file))
            cf.apply_keymap()
            return cf
        except toml.TomlDecodeError as e:
            raise Exception(f"Error parsing {file}: {e}")
    config = configparser.ConfigParser()
    config.read(file)
    cf = ConfigFile(file, {k: v for k, v in config.items() if k != "DEFAULT"}, format="ini")
    cf.apply_keymap()
    return cf

def _get_full_config(profile:str|None=None) -> tuple[Dict[str,Any], str|None]:
    """
    Returns a dictionary representing the props to be used for the currently active profile. Incorporates the following rules:
    1. The `snowflake_connection` key fills in the properties from the Snowflake connections.toml file.
    2. Profiles with the same name are merged in order of appearance, with earlier profiles taking precedence.
    3. The profile indicated by the `active_profile` key is merged into other top-level properties.
    """
    files = list(_find_config_file())
    file_path = first(files)
    if not files:
        return {}, file_path
    config_files = [
        _parse_and_map_config(file).filled_from_snowflake_connection()
        for file in files
    ]
    for i in range(len(config_files) - 1, 0, -1):
        config_files[i-1].merge(config_files[i])
    if not config_files:
        return {}, file_path
    root_file = config_files[0]
    if profile:
        root_file.config["active_profile"] = profile
    if ("active_profile" in root_file.config and
        root_file.config["active_profile"] in root_file.profiles):
        active_profile = root_file.config["active_profile"]
        config = {
            **root_file.profiles[active_profile],
            **root_file.config,
        }
    else:
        config = root_file.config
    return config, file_path

def has_platform(cfg, platform):
    return any(profile.get("platform") == platform for profile in cfg.get_combined_profiles().values())

def _legacy_config_files():
    """
    Generates all legacy config files found

    Includes both ~/.rai/config and rai.config (in a parent directory
    or the user's home directory)
    """
    yield from _search_upwards_for_file("rai.config")
    yield from _search_userdir_for_file(".rai.config")
    path = os.path.expanduser("~/.rai/config")
    if os.path.exists(path):
        yield path

def legacy_config_exists():
    return any(_legacy_config_files())

def all_configs_including_legacy():
    for file in _find_config_file():
        yield _parse_and_map_config(file)
    for file in _legacy_config_files():
        yield _parse_and_map_config(file)

def get_from_snowflake_connections_toml():
    user_config_path = os.path.expanduser("~/.snowflake/connections.toml")
    if os.path.exists(user_config_path):
        try:
            snow_config = toml.load(user_config_path)
        except toml.TomlDecodeError as e:
            raise Exception(f"Error parsing {user_config_path}: {e}")
        config = {}
        for profile in snow_config:
            config[profile] = {}
            for key in snow_config[profile]:
                config[profile][key] = snow_config[profile][key]
        return config

def to_rai_config(data:Dict[str, Any]) -> Dict[str, Any]:
    creds = config._read_client_credentials(data)
    _keys = ["host", "port", "region", "scheme", "audience"]
    result = {k: v for k, v in data.items() if k in _keys}
    result["credentials"] = creds
    return result


#--------------------------------------------------
# Config
#--------------------------------------------------

class Config():
    def __init__(self, profile:str|Dict[str,Any]|None=None, fetch=True):
        supplied_props = None
        if isinstance(profile, dict):
            supplied_props = profile
            profile = "__top_level__"
        if fetch:
            self.fetch(profile)
        else:
            self.profile = "__top_level__"
            self.props = {}
        self.cache_config_file_contents()
        if supplied_props is not None:
            for k, v in supplied_props.items():
                self.set(k, v)
            if not self.file_path:
                self.file_path = "__inline__"

    def get_profiles(self):
        if not self.original_toml:
            return {}
        return self.original_toml.get("profile", {})

    def fetch(self, profile:str|None=None):
        profile = profile or os.environ.get("RAI_PROFILE", None)
        cfg, path = _get_full_config(profile)
        self.profile = profile or cfg.get("active_profile", "__top_level__")
        self.file_path = path
        self.props = cfg

    def cache_config_file_contents(self):
        self.original_toml = tomlkit.document()
        if os.path.exists("raiconfig.toml"):
            with open("raiconfig.toml", "r") as f:
                try:
                    content = f.read()
                    if not content.endswith("\n"):
                        content += "\n"
                    self.original_toml = tomlkit.parse(content)
                except toml.TomlDecodeError as e:
                    raise Exception(f"Error parsing {self.file_path}: {e}")
            return

    def clone_profile(self):
        self.props = {k: v for k,v in self.props.items()}

    def get(self, name:str, default:Any|None=None, strict:bool=True):
        parts = name.split(".")
        props = self.props

        for part in parts[:-1]:
            props = props.get(part)
            if props is None:
                props = {}
                break

        val = props.get(parts[-1], os.environ.get(name, default))
        if val is None and strict:
            raise Exception(f"Missing config value for '{name}'")
        return val

    def set(self, name:str, value:str|int):
        self.props[name] = value

    def clear_props(self):
        self.props = {}

    def unset(self, name:str):
        del self.props[name]
    
    def update(self, other, save_progress=False):
        self.props.update(other)

    def __contains__(self, key):
        return key in self.props

    def items(self):
        return self.props.items()

    def print(self):
        for k, v in self.items():
            print(f"{k}: {v}")

    def to_rai_config(self) -> Dict[str, Any]:
        return to_rai_config(self.props)

    def save(self):
        new_document = tomlkit.document()

        # handle the active_profile key that appears at the top:
        if self.profile != "__top_level__":
            active_profile = self.profile
        else:
            active_profile = self.get("active_profile", None, strict=False)
        if active_profile is not None:
            new_document["active_profile"] = active_profile

        # add key-value pairs at the top level if this is the top-level profile:
        if self.profile == "__top_level__":
            for k, v in self.props.items():
                if k != "active_profile":
                    new_document[k] = v
        
        active_profile_table = tomlkit.table()
        for k, v in self.props.items():
            active_profile_table[k] = v

        original_top_level_profile = None

        profiles = tomlkit.table(is_super_table=True)

        def adjust_trailing_newlines(item, action):
            match item.__class__.__name__:
                case "TOMLDocument":
                    body = getattr(item, "body", None)
                    if body is None or len(body) == 0:
                        return                    
                    adjust_trailing_newlines(body[-1], action)
                case "tuple":
                    if len(item) < 2:
                        return
                    _, value = item
                    adjust_trailing_newlines(value, action)
                case "Container":
                    body = getattr(item, "body", None)
                    if body is None or len(body) == 0:
                        return
                    adjust_trailing_newlines(body[-1], action)
                case "Table":
                    value = getattr(item, "value", None)
                    if value is None:
                        return
                    adjust_trailing_newlines(value, action)
                case "Whitespace":
                    action(item)

        def ensure_profiles_added():
            if not new_document.get("profile"):
                # this hack is necessary because tomlkit sometimes
                # inserts spurious newlines just before the profile
                # section
                # related: https://github.com/sdispater/tomlkit/issues/48
                def trim_one(item):
                    if item.value.endswith("\n"):
                        item._s = item.value[:-1]
                adjust_trailing_newlines(new_document, trim_one)
                new_document.add("profile", profiles)

        # fill in from the original TOML file contents:
        for key, value in self.original_toml.items():
            # the "active-profile" check is here to make the migration smooth
            # from when we used "active-profile" instead of "active_profile"
            if key == "profile":
                ensure_profiles_added()
                for profile in value:
                    if profile == active_profile:
                        profiles.add(profile, active_profile_table)
                    else:
                        profiles.add(profile, value[profile])
            if key in new_document or key == "active_profile":
                continue
            elif key in all_prop_keys:
                ensure_profiles_added()
                if not original_top_level_profile:
                    k = 0
                    while True:
                        original = "original" if k == 0 else f"original-{k}"
                        if self.original_toml.get("profile", {}).get(original) is None:
                            break
                        k += 1
                    original_top_level_profile = original
                    profiles.add(original_top_level_profile, tomlkit.table())
                original_profile = cast(Table, profiles.get(original_top_level_profile))
                original_profile.add(key, value)
            else:
                # preserve all keys unrelated to the profiles:
                new_document[key] = value

        if active_profile is not None:
            ensure_profiles_added()
            if profiles.get(active_profile) is None:
                profiles.add(active_profile, active_profile_table)

        with open("raiconfig.toml", "w") as f:
            f.write(new_document.as_string())

    def _fill_in_with_defaults(self, defaults: Dict[str, Any], **kwargs):
        props = {k: v for k, v in kwargs.items() if k in defaults}
        self.update({
            **defaults,
            **self.props,
            **props,
        }, save_progress=True)

    def fill_in_with_azure_defaults(self, **kwargs):
        self._fill_in_with_defaults(azure_default_props, **kwargs)

    def fill_in_with_snowflake_defaults(self, **kwargs):
        self._fill_in_with_defaults(snowflake_default_props, **kwargs)

    def fill_in_with_defaults(self):
        platform = self.get("platform", None, strict=False)
        if platform == "azure":
            self.fill_in_with_azure_defaults()
        elif platform == "snowflake":
            self.fill_in_with_snowflake_defaults()
        else:
            self.set("platform", "snowflake")
            self.fill_in_with_snowflake_defaults()


from pathlib import Path

import pytest
from gentest.exec import path_to_slug, validate_query_results
from relationalai.clients import config as cfg

test_case_dir = Path(__file__).parent / "test_cases"
test_case_files = [path for path in test_case_dir.iterdir() if path.suffix == ".py"]

@pytest.mark.parametrize("file_path", test_case_files, ids=lambda path: path_to_slug(path, test_case_dir))
def test_snapshots(file_path: Path, snapshot, engine_config: cfg.Config):
    validate_query_results(file_path, snapshot, {"test_config": engine_config})



from relationalai.metamodel import Type

Schema = dict[str, Type]

LoadType = str

class UnsupportedTypeError(Exception):
    def __init__(self, message: str, type: str|None = None):
        super().__init__(message)
        self.type = type



import relationalai as rai
from relationalai.std.graphs import Graph

# Test case taken from:
# https://github.com/RelationalAI/raicode/blob/8da410b949e8c47913a6faf4809234928fd8c238/src/rel/graph.rel#L537

model = rai.Model("indegreeTest", config=globals()["test_config"])
Node = model.Type("Node")

edge_list = [(11, 12), (12, 13), (13, 13), (12, 14)]
with model.rule(dynamic=True):
    for (x, y) in edge_list:
        node1 = Node.add(id=x)
        node2 = Node.add(id=y)
        node1.set(adjacent_to=node2)

# Undirected case
undirected_graph = Graph(model, undirected=True)
undirected_graph.edges.extend(Node.adjacent_to)

with model.query() as select:
    node = Node()
    degree = undirected_graph.compute.outdegree(node)
    response = select(node.id, degree)

# Directed case
directed_graph = Graph(model)
directed_graph.edges.extend(Node.adjacent_to)

with model.query() as select:
    node = Node()
    degree = directed_graph.compute.outdegree(node)
    response = select(node.id, degree)



# pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.std.graphs import Graph
from relationalai.std.aggregates import sum
import rich

#--------------------------------------------------
# Types
#--------------------------------------------------

model = rai.Model("MyCoolDatabase")
Person = model.Type("Person")
Criminal = model.Type("Criminal")
Transaction = model.Type("Transaction")

#--------------------------------------------------
# Load data
#--------------------------------------------------

with model.rule():
    joe = Person.add(name="Joe")
    jane = Person.add(name="Jane")
    bob = Person.add(name="Bob")
    dylan = Person.add(Criminal, name="Dylan")

    joe.set(knows=jane)
    jane.set(knows=bob)
    bob.set(knows=joe)
    bob.set(knows=dylan)
    dylan.set(knows=joe)

    Transaction.add(from_=joe, to=jane, amount=100)
    Transaction.add(from_=joe, to=jane, amount=1000)
    Transaction.add(from_=joe, to=jane, amount=10)
    Transaction.add(from_=joe, to=bob, amount=10)

#--------------------------------------------------
# Graph
#--------------------------------------------------

graph = Graph(model)
nodes, edges = graph.nodes, graph.edges

nodes.extend(Person, label=Person.name)
nodes.extend(Criminal, criminal=True)
edges.extend(Person.knows, label="knows")

with model.rule():
    t = Transaction()
    weight = sum(t, t.amount, per=[t.from_, t.to])
    edges.add(t.from_, t.to, label="Transfer", weight=weight)

with model.rule():
    p = Person()
    rank = graph.compute.pagerank(p)
    nodes.add(p, rank=rank * 5)

#--------------------------------------------------
# Paths
#--------------------------------------------------

# Transfer = Edge(Transaction, Transaction.from_, Transaction.to)
# Teller = Edge(Transaction, Transaction.from_, Transaction.teller)

# with model.query() as select:
#     p = Path(Person, Transfer, Criminal)
#     select(p)

# with model.query() as select:
#     person = Person(name="John")
#     t = Transfer(); t.amount > 10000
#     p = Path(person, t, Criminal)
#     select(p)

# with model.query() as select:
#     p = Path(Person, Transfer[0:2], Criminal)
#     select(p)

# with model.query() as select:
#     p = Path(Person, Transfer, Person, Knows, Criminal)[1:2]
#     select(p)

# with model.query() as select:
#     t = Transfer(); t.amount > 10000
#     sub = Path(Person, t, Person)
#     p = Path(Person, sub[1:3], Criminal)
#     select(p[0])

#--------------------------------------------------
# Notes
#--------------------------------------------------
"""

Extends is a convenient way to take an existing set and add it to the graph, but
arbitrary logic can be used with nodes.add and edges.add to have explicit control:

    with model.rule():
        p = Person()
        nodes.add(p, label=p.name)
        with Criminal(p):
            nodes.add(p, color="red")
        edges.add(p, p.knows, color="#aaa")

If we were in SPCS, the data would come from snowflake tables which likely represent
hyperedges and so we might have N logical edges coming from a single table:

    Transfer = Edge(Transaction, Transaction.from_, Transaction.to)
    Teller = Edge(Transaction, Transaction.from_, Transaction.teller)

    edges.extend(Transfer)
    edges.extend(Teller)
"""

#--------------------------------------------------
# Go
#--------------------------------------------------

data = graph.fetch()
rich.print(data)
graph.visualize(three=True, style={
    "node": {
        "color": lambda x: "red" if x.get("criminal") else "blue",
        "size": lambda x: (x["rank"] * 2) ** 2
    },
    "edge": {
        "color": "yellow",
    }
}).display()


from typing import Any, Iterable, Sequence

import pandas as pd
from railib import api

from relationalai.dsl import safe_symbol
from relationalai.errors import RelQueryError

#-------------------------------------------------------------------------------
# Emitting
#-------------------------------------------------------------------------------

def emit_literal(v: Any):
    """Emit `v` as its equivalent literal representation in rel."""
    if isinstance(v, str):
        return f'"{v}"'
    if isinstance(v, tuple):
        return ", ".join(emit_literal(item) for item in v)
    return v

def emit_nested_relation(prefix: str, obj: dict|None = None, keys: Iterable[str]|None = None, raw = False) -> str:
    """Emit a set of defs encoding `obj` in GNF on `prefix`."""
    obj = obj or {}
    result = ""
    for k in keys or obj.keys():
        v = obj.get(k, None)
        if v is not None:
            result += f"def {prefix}[{safe_symbol(k)}] = {emit_literal(v) if not raw else v}\n"
    return result

#-------------------------------------------------------------------------------
# Result Handling
#-------------------------------------------------------------------------------

_known_problems: list[Any] = []
def ignore_known_problems(problems: Sequence[Any]):
    """Filter out issues already in `problems` when checking responses for rel problems."""
    global _known_problems
    _known_problems.clear()
    _known_problems.extend(problems)

def assert_no_problems(res: api.TransactionAsyncResponse):
    """Throw a vaguely-readable exception if rel problems were reported with the given transaction."""
    new_problems = []
    for problem in res.problems:
        if problem not in _known_problems:
            new_problems.append(problem)

    if new_problems:
        raise RelQueryError(new_problems)

def maybe_scalarize(v):
    if getattr(v, "__len__", None) and len(v) == 1:
        return v[0]
    return v

def process_gnf_results(gnf: Sequence, *key_names: str):
    """Process GNF results into a nested object keyed by the key(s) in `key_names`"""
    assert len(key_names) > 0, "Must supply a name for the GNF key(s)"
    key_size = len(key_names)

    obj = {}

    for rel in gnf:
        sig = rel["relationId"].split("/")
        if sig[1] != ":output":
            continue

        table: pd.DataFrame = rel["table"].to_pandas()

        key_cols = table.columns[0:key_size]

        for _, row in table.iterrows():
            raw_keys = tuple(row[col] for col in key_cols)
            keys = maybe_scalarize(raw_keys)
            vals = [row[col] for col in table.columns[key_size:]]

            entry = obj.setdefault(keys, {k: v for k, v in zip(key_names, raw_keys)})

            cur = 4
            field = sig[cur - 2][1:]
            while len(sig) > cur and sig[cur][0] == ":":
                entry = entry.setdefault(field, {})
                field = sig[cur][1:]
                cur += 2
            entry[field] = maybe_scalarize(vals)

    return obj



import ast
from dataclasses import dataclass
import inspect
import json
from textwrap import dedent
import os
from typing import Dict
import logging

from pandas import DataFrame
from .errors import Errors
from .metamodel import Action, Task

DEBUG = True

#--------------------------------------------------
# Logging
#--------------------------------------------------

class JsonFormatter(logging.Formatter):
    def format(self, record):
        return json.dumps(record.msg)

class FlushingFileHandler(logging.FileHandler):
    def emit(self, record):
        super().emit(record)
        self.flush()

with open('debug.jsonl', 'w'):
    pass

logger = logging.getLogger("pyrellogger")
logger.setLevel(logging.DEBUG)
file_handler = FlushingFileHandler('debug.jsonl', mode='a')
file_handler.setFormatter(JsonFormatter())
logger.addHandler(file_handler)
logger.propagate = False

def time(type:str, elapsed:float, results:DataFrame = DataFrame(), **kwargs):
    if DEBUG:
        d = {"event":"time",
             "type": type,
             "elapsed": elapsed,
             "results": {"values": results.head(20).to_json(orient="records"),
                          "count": len(results)}
            }
        d.update(kwargs)
        logger.debug(d)

def handle_compilation(compilation):
    if DEBUG:
        passes = [{"name": p[0], "task": str(p[1]), "elapsed": p[2]} for p in compilation.passes]
        (file, line, block) = compilation.get_source()
        emitted = compilation.emitted
        if isinstance(emitted, list):
            emitted = "\n\n".join(emitted)
        logger.debug({
            "event":"compilation",
            "source": {"file": file, "line": line, "block": block},
            "task": str(compilation.task),
            "passes": passes,
            "emitted": emitted,
            "emit_time": compilation.emit_time
        })

#--------------------------------------------------
# SourceInfo
#--------------------------------------------------

@dataclass
class SourceInfo:
    file: str = "Unknown"
    line: int = 0
    source: str = ""
    block: ast.AST|None = None

    def modify(self, transformer:ast.NodeTransformer):
        if not self.block:
            raise Exception("Cannot modify source info without a block")

        new_block = transformer.visit(self.block)
        return SourceInfo(self.file, self.line, ast.unparse(new_block), new_block)

#--------------------------------------------------
# Jupyter
#--------------------------------------------------

class Jupyter:
    def __init__(self):
        self.dirty_cells = set()
        try:
            from IPython import get_ipython # type: ignore
            self.ipython = get_ipython()
            if self.ipython:
                self.ipython.events.register('pre_run_cell', self.pre_run_cell)
                self.dirty_cells.add(self.cell())
        except ImportError:
            self.ipython = None

    def pre_run_cell(self, info):
        self.dirty_cells.add(info.cell_id)

    def cell_content(self):
        if self.ipython:
            last_input = self.ipython.user_ns['In'][-1]
            return (last_input, f"In[{len(self.ipython.user_ns['In'])}]")
        return ("", "")

    def is_colab(self):
        if not self.ipython:
            return False
        try:
            parent = self.ipython.get_parent() #type: ignore
        except Exception:
            return False
        if not parent or "metadata" not in parent:
            return False
        return "colab" in parent["metadata"]

    def cell(self):
        if self.ipython:
            if self.is_colab():
                return self.ipython.get_parent()["metadata"]["colab"]["cell_id"] #type: ignore
            else:
                try:
                    return self.ipython.get_parent()["metadata"]["cellId"] #type: ignore
                except Exception:
                    return None
        return None

jupyter = Jupyter()

#--------------------------------------------------
# Position capture
#--------------------------------------------------

def first_non_relationalai_frame(frame):
    while frame and frame.f_back:
        file = frame.f_code.co_filename
        if "site-packages/relationalai" not in file and "src/relationalai" not in file:
            break
        frame = frame.f_back
    return frame

def capture_code_info(steps=None):
    # Get the current frame and go back to the caller's frame
    caller_frame = inspect.currentframe()
    if steps is not None:
        for _ in range(steps):
            if not caller_frame or not caller_frame.f_back:
                break
            caller_frame = caller_frame.f_back
    else:
        caller_frame = first_non_relationalai_frame(caller_frame)

    if not caller_frame:
        return

    caller_filename = caller_frame.f_code.co_filename
    caller_line = caller_frame.f_lineno

    relative_filename = os.path.relpath(caller_filename, os.getcwd())

    # Read the source code from the caller's file
    source_code = None
    try:
        with open(caller_filename, "r") as f:
            source_code = f.read()
    except IOError:
        (jupyter_code, jupyter_cell) = jupyter.cell_content()
        if jupyter_code:
            source_code = jupyter_code
            relative_filename = jupyter_cell

    if not source_code:
        return SourceInfo(relative_filename, caller_line)
    else:
        return find_block_in(source_code, caller_line, relative_filename)

def find_block_in(source_code: str, caller_line: int, relative_filename: str):
    # Parse the source code into an AST
    tree = ast.parse(source_code)

    # Find the node that corresponds to the call
    class BlockFinder(ast.NodeVisitor):
        def __init__(self, target_lineno):
            self.target_lineno = target_lineno
            self.block_node = None

        def generic_visit(self, node):
            if hasattr(node, "lineno") and node.lineno == self.target_lineno:
                self.block_node = node
                # Stop visiting once the target node is found
                return
            ast.NodeVisitor.generic_visit(self, node)

    finder = BlockFinder(caller_line)
    finder.visit(tree)

    if finder.block_node:
        # Extract the lines from the source code
        start_line = finder.block_node.lineno
        end_line = getattr(finder.block_node, "end_lineno", start_line)

        block_lines = source_code.splitlines()[start_line - 1:end_line]
        block_code = "\n".join(block_lines)
        return SourceInfo(relative_filename, caller_line, dedent(block_code), finder.block_node)

    lines = source_code.splitlines()
    if caller_line > len(lines):
        return SourceInfo(relative_filename, caller_line)
    return SourceInfo(relative_filename, caller_line, lines[caller_line - 1])

def check_errors(task:Task|Action):
    class ErrorFinder(ast.NodeVisitor):
        def __init__(self, start_line):
            self.errors = []
            self.start_line = start_line

        def to_line_numbers(self, node):
            return (node.lineno, node.end_lineno)

        def generic_visit(self, node):
            if isinstance(node, ast.If):
                Errors.invalid_if(task, *self.to_line_numbers(node))
            elif isinstance(node, ast.For) or isinstance(node, ast.While):
                Errors.invalid_loop(task, *self.to_line_numbers(node))
            elif isinstance(node, ast.Try):
                Errors.invalid_try(task, *self.to_line_numbers(node))

            ast.NodeVisitor.generic_visit(self, node)

    source = get_source(task)
    if not source or not source.block:
        return
    ErrorFinder(source.line).visit(source.block)


sources:Dict[Task|Action, SourceInfo|None] = {}
def set_source(item, steps=1, dynamic=False):
    found = capture_code_info()
    if found:
        sources[item] = found
        if not dynamic:
            check_errors(item)
            pass
    return found

def get_source(item):
    return sources.get(item)



# pyright: reportUnusedExpression=false
from collections import defaultdict
import json
import textwrap
from railib.api import TransactionAsyncResponse, poll_with_specified_overhead, _parse_metadata_proto
import requests
import snowflake.connector
import pyarrow as pa
import pandas as pd
import time

from .. import debugging

from typing import Any, Dict, Iterable, Tuple, List, cast
import base64

from pandas import DataFrame

from relationalai.clients.azure import MILLISECONDS_PER_DAY, UNIXEPOCH

from ..tools.cli_controls import Spinner

from ..clients.types import AvailableModel, Import, ImportSource, ImportSourceTable
from ..clients.config import Config
from ..clients.client import Client, ResourceProvider
from .. import dsl, rel, metamodel as m
from ..errors import Errors
from .. import std
import re

USE_EXEC_ASYNC = True

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def type_to_sql(type) -> str:
    if type is str:
        return "VARCHAR"
    if type is int:
        return "NUMBER"
    if type is float:
        return "FLOAT"
    if type is bool:
        return "BOOLEAN"
    if type is dict:
        return "VARIANT"
    if type is list:
        return "ARRAY"
    if type is bytes:
        return "BINARY"
    if isinstance(type, dsl.Type):
        return "NUMBER"
    raise ValueError(f"Unknown type {type}")

def type_to_snowpark(type) -> str:
    if type is str:
        return "StringType"
    if type is int:
        return "IntegerType"
    if type is float:
        return "FloatType"
    if type is bool:
        return "BooleanType"
    if type is dict:
        return "MapType"
    if type is list:
        return "ArrayType"
    if type is bytes:
        return "BinaryType"
    if isinstance(type, dsl.Type):
        return "IntegerType"
    raise ValueError(f"Unknown type {type}")

#--------------------------------------------------
# Resources
#--------------------------------------------------

APP_NAME = "___RAI_APP___"

class Resources(ResourceProvider):
    def __init__(self, profile:str|None=None, config:Config|None=None):
        super().__init__(profile, config=config)
        self._conn = None
        self._app_name = "relationalai"
        self._pending_transactions: list[str] = []

    def _exec(self, code:str):
        if not self._conn:
            self._conn = snowflake.connector.connect(
                user=self.config.get('user'),
                password=self.config.get('password'),
                account=self.config.get('account'),
                warehouse=self.config.get('warehouse', ""),
                # database=self.config.get('database', ""),
                # schema=self.config.get('schema', ""),
                role=self.config.get('role', ""),
                client_store_temporary_credential=True,
                client_request_mfa_token=True,
            )
            self._app_name = self.config.get('rai_app_name', "relationalai")
        try:
            return self._conn.cursor().execute(code.replace(APP_NAME, self._app_name))
        except Exception as e:
            if re.search(r"Database '(.+)' does not exist or not authorized\.", str(e)):
                print("\n")
                name = re.search(r"Database '(.+)' does not exist or not authorized\.", str(e)).group(1) #type: ignore
                Errors.snowflake_app_missing(name)
            else:
                raise e


    def reset(self):
        self._conn = None

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------

    def list_engines(self):
        results = self._exec(f"select NAME, SIZE, STATUS from {APP_NAME}.api.engines")
        if not results:
            return []
        return [{"name":name, "size":size, "state":state}
                for (name, size, state) in results.fetchall()]

    def create_engine(self, name:str, size:str, pool:str = ""):
        if not pool:
            raise ValueError("Pool is required")
        try:
            self._exec(f"call {APP_NAME}.api.create_engine('{name}', '{pool}', '{size}');")
        except Exception as e:
            raise Exception("Failed to create engine") from e

    def delete_engine(self, name:str):
        self._exec(f"call {APP_NAME}.api.delete_engine('{name}');")

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    def list_graphs(self) -> List[AvailableModel]:
        results = self._exec(f"select NAME from {APP_NAME}.api.databases WHERE state <> 'DELETED'")
        if not results:
            return []
        return [{"name": name} for (name,) in results.fetchall()]

    def get_graph(self, name:str):
        results = self._exec(f"""
            select ID, NAME, CREATED_ON, DELETED_ON, STATE
            from {APP_NAME}.api.databases
            where name='{name}' AND state <> 'DELETED'
        """)
        if not results:
            return None
        return results.fetchone()

    def create_graph(self, name: str):
        self._exec(f"call {APP_NAME}.api.create_database('{name}');")

    def delete_graph(self, name:str):
        self._exec(f"call {APP_NAME}.api.delete_database('{name}');")

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        pass

    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        lines = []
        for (name, code) in models:
            name = name.replace("\"", "\\\"")
            assert "\"\"\"\"\"\"\"" not in code, "Code literals must use fewer than 7 quotes."

            lines.append(textwrap.dedent(f"""
            def delete:rel:catalog:model["{name}"] = rel:catalog:model["{name}"]
            def insert:rel:catalog:model["{name}"] = raw\"\"\"\"\"\"\"{code}\"\"\"\"\"\"\"
            """))
        rel_code = "\n\n".join(lines)
        self.exec_raw(database, engine, rel_code, readonly=False)
        # TODO: handle SPCS errors once they're figured out
        return []

    def delete_model(self, database:str, engine:str, name:str):
        self.exec_raw(database, engine, f"def delete:rel:catalog:model[\"{name}\"] = rel:catalog:model[\"{name}\"]", readonly=False)

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        return []

    def create_export(self, database: str, engine: str, func_name: str, inputs: List[Tuple[str, str, Any]], out_fields: List[Tuple[str, Any]], code: str):
        sql_inputs = ", ".join([f"{name} {type_to_sql(type)}" for (name, var, type) in inputs])
        sql_out = ", ".join([f"{name} {type_to_sql(type)}" for (name, type) in out_fields])
        py_outs = ", ".join([f"StructField(\"{name}\", {type_to_snowpark(type)}())" for (name, type) in out_fields])
        py_inputs = ", ".join([name for (name, *rest) in inputs])
        safe_rel = code.replace("'", "\\'").replace("{", "{{").replace("}", "}}").strip()
        for (name, var, type) in inputs:
            if type is str:
                safe_rel = safe_rel.replace(var, f"'{{{name}.replace(\"'\", \"\\'\")}}'")
            else:
                safe_rel = safe_rel.replace(var, f"{{{name}}}")
        safe_rel = safe_rel.replace("\n", "\\n")
        if py_inputs:
            py_inputs = f", {py_inputs}"
        sql_code = textwrap.dedent(f"""
                CREATE OR REPLACE PROCEDURE {func_name}({sql_inputs})
                RETURNS TABLE({sql_out})
                LANGUAGE PYTHON
                RUNTIME_VERSION = '3.8'
                PACKAGES = ('snowflake-snowpark-python')
                HANDLER = 'handle'
                AS
                $$
                import json
                from snowflake.snowpark import DataFrame
                from snowflake.snowpark.functions import col
                from snowflake.snowpark.types import StringType, IntegerType, StructField, StructType, FloatType, MapType, ArrayType, BooleanType, BinaryType

                def handle(session{py_inputs}):
                    rel_code = f"{safe_rel}"
                    results = session.sql(f"select {APP_NAME}.api.exec('{database}','{engine}','{{rel_code}}', true);")
                    table_results = []
                    for row in results.collect():
                        parsed = json.loads(row[0])
                        table_results.extend(parsed["data"])

                    schema = StructType([{py_outs}])
                    return session.create_dataframe(table_results, schema=schema)
                $$;""")
        start = time.perf_counter()
        self._exec(sql_code)
        debugging.time("export", time.perf_counter() - start, DataFrame(), code=sql_code)
        return

    def delete_export(self, model: str, engine: str, export: str):
        pass

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    def list_imports(self, model:str) -> list[Import]:
        results = self._exec(f"""
            select FQ_OBJECT_NAME
            from {APP_NAME}.api.data_streams
            where RAI_DATABASE='{model}';
        """)
        if not results:
            return []
        return [cast(Import, {"name": name, "type": "snowflake_table"}) for (name,) in results.fetchall()]

    def create_import_stream(self, source:ImportSource, model:str, rate = 1, options: dict|None = None):
        assert isinstance(source, ImportSourceTable), "Snowflake integration only supports loading from SF Tables right now. Try loading your data as a table via the snowflake interface first."
        object = source.fqn

        if object.lower() in [x["name"].lower() for x in self.list_imports(model)]:
            return

        try:
            self._exec(f"call {APP_NAME}.api.setup_cdc('{self.config.get('engine')}');")
        except Exception:
           pass

        try:
            self._exec(f"ALTER TABLE {object} SET CHANGE_TRACKING = TRUE;")
        except Exception:
            pass

        self._exec(f"""call {APP_NAME}.api.create_data_stream(
             {APP_NAME}.api.object_reference('TABLE', '{object}'),
             '{model}',
             '{object.replace('.', '_').lower()}');""")
        return

    def create_import_snapshot(self, source:ImportSource, model:str, _: dict|None = None):
        raise Exception("Snowflake integration doesn't support snapshot imports yet")

    def delete_import(self, import_name:str, model:str):
        self._exec(f"""call {APP_NAME}.api.delete_data_stream(
             '{import_name}',
             '{model}'
        );""")
        return

    #--------------------------------------------------
    # Exec Sync
    #--------------------------------------------------

    def _exec_sync_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs = {}):
        if inputs:
            raise Exception("Inputs aren't currently supported using exec_sync in SPCS.")

        return self._exec(f"select {APP_NAME}.api.exec('{database}','{engine}','{raw_code}', {readonly});")

    def _format_results_sync(self, results, task:m.Task) -> Tuple[DataFrame, List[Any]]:
        parsed_results = []
        parsed_problems = []
        if results:
            for row in results:
                parsed = json.loads(row[0])
                if parsed.get("problems"):
                    for problem in parsed["problems"]:
                        parsed_problems.append(problem)
                else:
                    parsed_results.extend(parsed["data"])
        try:
            data_frame = DataFrame(parsed_results, columns=task.return_cols())
        except Exception:
            data_frame = DataFrame(parsed_results)
        return (data_frame, parsed_problems)

    #--------------------------------------------------
    # Exec Async
    #--------------------------------------------------

    def _check_exec_async_status(self, txn_id: str):
        """Check whether the given transaction has completed."""
        response = self._exec(f"CALL {APP_NAME}.api.get_transaction('{txn_id}');")
        assert response, f"No results from get_transaction('{txn_id}')"
        status: str = next(iter(response))[2]

        if status == "ABORTED":
            raise Exception(f"Transaction aborted while waiting for results '{txn_id}'")

        return status == "COMPLETED"

    def _list_exec_async_artifacts(self, txn_id: str) -> Dict[str, str]:
        """Grab the list of artifacts produced in the transaction and the URLs to retrieve their contents."""
        response = self._exec(f"CALL {APP_NAME}.api.list_transaction_outputs('{txn_id}');")
        assert response, f"No results from list_transaction_outputs('{txn_id}')"
        return {name: url for name, url in response}

    def _fetch_exec_async_artifacts(self, artifact_urls: Dict[str, str]) -> Dict[str, Any]:
        """Grab the contents of the given artifacts from SF."""
        contents = {}
        with requests.Session() as session:
            for name, url in artifact_urls.items():
                response = session.get(url)
                response.raise_for_status() # throw if something goes wrong.
                if name.endswith(".json"):
                    contents[name] = response.json()
                else:
                    contents[name] = response.content

        return contents

    def _parse_exec_async_results(self, arrow_files: List[Tuple[str, bytes]]):
        """Mimics the logic in _parse_arrow_results of railib/api.py#L303 without requiring a wrapping multipart form."""
        results = []

        for file_name, file_content in arrow_files:
            with pa.ipc.open_stream(file_content) as reader:
                schema = reader.schema
                batches = [batch for batch in reader]
                table = pa.Table.from_batches(batches=batches, schema=schema)
                results.append({"relationId": file_name, "table": table})

        return results

    def _exec_async_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs={}):
        if inputs:
            raise Exception("Inputs aren't currently supported using exec_async in SPCS.")

        response = self._exec(f"CALL {APP_NAME}.api.exec_async('{database}','{engine}','{raw_code}', {readonly});")
        if not response:
            raise Exception("No results from exec_async")

        # Grab the txn_id from the response
        txn_id, status = next(iter(response))
        # Wait for completion or failure
        if status != "COMPLETED":
            self._pending_transactions.append(txn_id)
            poll_with_specified_overhead(lambda: self._check_exec_async_status(txn_id), 0.2)

        self._pending_transactions.remove(txn_id)

        # List the result artifacts (and the URLs to retrieve them)
        artifact_urls = self._list_exec_async_artifacts(txn_id)
        # Actually retrieve them
        artifacts = self._fetch_exec_async_artifacts(artifact_urls)

        meta = _parse_metadata_proto(artifacts["metadata.proto"])
        meta_json = artifacts["metadata.json"]

        # We use the metadata to map arrow files to the relations they contain data for.
        # In Azure, this is provided sideband via response headers.
        arrow_files_to_relations = {}
        for ix, relation in enumerate(cast(Any, meta).relations):

            arrow_file = relation.file_name
            relation_id = meta_json[ix]["relationId"]
            arrow_files_to_relations[arrow_file] = relation_id

        # Hydrate the arrow files into tables
        results = self._parse_exec_async_results([(arrow_files_to_relations[name], content) for name, content in artifacts.items() if name.endswith(".arrow")])

        rsp = TransactionAsyncResponse()
        # @FIXME: Hardcoding the `state` feels somewhat unsafe, but we can't have gotten here in the code otherwise.
        # @FIXME: Missing `response_format_version`, which isn't obviously present anywhere in the results.
        rsp.transaction = {"id": txn_id, "state": "COMPLETED", "response_format_version": None}
        rsp.metadata = meta
        rsp.problems = artifacts["problems.json"]
        rsp.results = results
        return rsp

    # Copied directly from azure.py
    def _has_errors(self, results):
        if len(results.problems):
            for problem in results.problems:
                if problem['is_error'] or problem['is_exception']:
                    return True

    # Copied directly from azure.py
    def _format_results_async(self, results, task:m.Task) -> Tuple[DataFrame, List[Any]]:
        data_frame = DataFrame()
        if len(results.results):
            for result in results.results:
                types = [t for t in result["relationId"].split("/") if t != "" and not t.startswith(":")]
                result_frame:DataFrame = result["table"].to_pandas()
                for i, col in enumerate(result_frame.columns):
                    if "UInt128" in types[i]:
                        result_frame[col] = result_frame[col].apply(lambda x: base64.b64encode(x.tobytes()).decode()[:-2])
                    if types[i] == "Dates.DateTime":
                        result_frame[col] = pd.to_datetime(result_frame[col] - UNIXEPOCH, unit="ms")
                    if types[i] == "Dates.Date":
                        result_frame[col] = pd.to_datetime(result_frame[col] * MILLISECONDS_PER_DAY - UNIXEPOCH, unit="ms")
                ret_cols = task.return_cols()
                if len(ret_cols) and len(result_frame.columns) == len(ret_cols):
                    result_frame.columns = task.return_cols()[0:len(result_frame.columns)]
                result["table"] = result_frame
                if ":output" in result["relationId"]:
                    data_frame = pd.concat([data_frame, result_frame], ignore_index=True)
        return (data_frame, results.problems)

    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    def exec_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs: dict = {}):
        raw_code = raw_code.replace("'", "\\'") # @NOTE: If collapsing to a single exec, make sure to copy this line into it.
        if USE_EXEC_ASYNC:
            return self._exec_async_raw(database, engine, raw_code, readonly, inputs=inputs)
        else:
            return self._exec_sync_raw(database, engine, raw_code, readonly, inputs=inputs)

    def format_results(self, results, task:m.Task) -> Tuple[DataFrame, List[Any]]:
        if USE_EXEC_ASYNC:
            return self._format_results_async(results, task)
        else:
            return self._format_results_sync(results, task)

    def cancel_pending_transactions(self):
        for txn_id in self._pending_transactions:
            self._exec(f"CALL {APP_NAME}.api.CANCEL_TRANSACTION('{txn_id}');")

    #--------------------------------------------------
    # Snowflake specific
    #--------------------------------------------------

    def list_warehouses(self):
        results = self._exec("SHOW WAREHOUSES")
        if not results:
            return []
        return [{"name":name}
                for (name, *rest) in results.fetchall()]

    def list_compute_pools(self):
        results = self._exec("SHOW COMPUTE POOLS")
        if not results:
            return []
        return [{"name":name}
                for (name, *rest) in results.fetchall()]

    def list_roles(self):
        results = self._exec("SELECT CURRENT_AVAILABLE_ROLES()")
        if not results:
            return []
        # the response is a single row with a single column containing
        # a stringified JSON array of role names:
        row = results.fetchone()
        if not row:
            return []
        return [{"name": name} for name in json.loads(row[0])]

    def list_apps(self):
        results = self._exec("SHOW APPLICATIONS")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_databases(self):
        results = self._exec("SHOW DATABASES")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_sf_schemas(self, database:str):
        results = self._exec(f"SHOW SCHEMAS IN {database}")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_tables(self, database:str, schema:str):
        results = self._exec(f"SHOW TABLES IN {database}.{schema}")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]


    def schema_info(self, database:str, schema:str, tables:Iterable[str]):
        pks = self._exec(f"SHOW PRIMARY KEYS IN SCHEMA {database}.{schema};")
        fks = self._exec(f"SHOW IMPORTED KEYS IN SCHEMA {database}.{schema};")
        tables = ", ".join([f"'{x.upper()}'" for x in tables])
        columns = self._exec(f"""
            select TABLE_NAME, COLUMN_NAME, DATA_TYPE
            from {database.upper()}.INFORMATION_SCHEMA.COLUMNS
            where TABLE_SCHEMA = '{schema.upper()}'
            and TABLE_NAME in ({tables})
            and TABLE_CATALOG = '{database.upper()}';
        """)
        results = defaultdict(lambda: {"pks": [], "fks": {}, "columns": {}})
        if pks:
            for row in pks:
                results[row[3].lower()]["pks"].append(row[4].lower()) # type: ignore
        if fks:
            for row in fks:
                results[row[7].lower()]["fks"][row[8].lower()] = row[3].lower()
        if columns:
            for row in columns:
                results[row[0].lower()]["columns"][row[1].lower()] = row[2].lower()
        return results

#--------------------------------------------------
# Snowflake Wrapper
#--------------------------------------------------

class PrimaryKey:
    pass

class Snowflake:
    def __init__(self, model, auto_import=False):
        self._model = model
        self._auto_import = auto_import
        if not isinstance(model._client.resources, Resources):
            raise ValueError("Snowflake model must be used with a snowflake config")
        self._dbs = {}
        imports = model._client.resources.list_imports(model.name)
        self._import_structure(imports)

    def _import_structure(self, imports: list[Import]):
        tree = self._dbs
        # pre-create existing imports
        schemas = set()
        for item in imports:
            database_name, schema_name, table_name = item["name"].lower().split('.')
            database = getattr(self, database_name)
            schema = getattr(database, schema_name)
            schemas.add(schema)
            schema._add(table_name, is_imported=True)
        for schema in schemas:
            schema._finalize()
        return tree

    def __getattribute__(self, __name: str) -> 'SnowflakeDB':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._dbs:
            return self._dbs[__name]
        self._dbs[__name] = SnowflakeDB(self, __name)
        return self._dbs[__name]

class SnowflakeDB:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._schemas = {}

    def __getattribute__(self, __name: str) -> 'SnowflakeSchema':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._schemas:
            return self._schemas[__name]
        self._schemas[__name] = SnowflakeSchema(self, __name)
        return self._schemas[__name]

class SnowflakeSchema:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._tables = {}

    def _finalize(self):
        self._table_info = self._model._client.resources.schema_info(self._parent._name, self._name, self._tables.keys())
        for table in self._tables.values():
            table._finalize()

    def _add(self, name, is_imported=False):
        name = name.lower()
        if name in self._tables:
            return self._tables[name]
        self._tables[name] = SnowflakeTable(self, name, is_imported=is_imported)
        return self._tables[name]

    def __getattribute__(self, __name: str) -> 'SnowflakeTable':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        return self._add(__name)

class SnowflakeTable(dsl.Type):
    def __init__(self, parent, name, is_imported=False):
        super().__init__(parent._model, f"sf_{name}", ["namespace", "fqname", "describe"])
        self._name = name
        self._model = parent._model
        self._parent = parent
        self._aliases = {}
        self._finalzed = False
        if not is_imported and self._parent._parent._parent._auto_import:
            with Spinner(f"Creating stream for {self.fqname()}", f"Stream created for {self.fqname()}"):
                self._model._client.resources.create_import_stream(self.fqname(), self._model.name)
            print("")
            parent._tables[name] = self
            parent._finalize()
        elif not is_imported:
            Errors.snowflake_import_missing(debugging.capture_code_info(4), self.fqname(), self._model.name)


    def _finalize(self):
        if self._finalzed:
            return

        self._finalzed = True
        self._schema = self._parent._table_info[self._name]
        relation_name = self.fqname().replace(".", "_").lower()
        model:dsl.Graph = self._model
        model.install_raw(f"bound {relation_name}")

        with model.rule(dynamic=True):
            prop, id, val = std.Vars(3)
            if self._schema["pks"]:
                getattr(getattr(std.rel, relation_name), self._schema["pks"][0].upper())(id, val)
            else:
                getattr(std.rel, relation_name)(prop, id, val)
            self.add(snowflake_id=id)

        for prop, prop_type in self._schema["columns"].items():
            with model.rule():
                id, val = std.Vars(2)
                getattr(getattr(std.rel, relation_name), prop.upper())(id, val)
                self(snowflake_id=id).set(**{prop.lower(): val})

    def namespace(self):
        return f"{self._parent._parent._name}.{self._parent._name}"

    def fqname(self):
        return f"{self.namespace()}.{self._name}"

    def describe(self, **kwargs):
        model = self._model
        for k, v in kwargs.items():
            if v is PrimaryKey:
                self._schema["pks"] = [k]
            elif isinstance(v, tuple):
                (table, name) = v
                if isinstance(table, SnowflakeTable):
                    fk_table = table
                    pk = fk_table._schema["pks"]
                    with model.rule():
                        inst = fk_table()
                        me = self()
                        getattr(inst, pk[0]) == getattr(me, k)
                        me.set(**{name: inst})
                else:
                    raise ValueError(f"Invalid foreign key {v}")
            else:
                raise ValueError(f"Invalid column {k}={v}")
        return self

#--------------------------------------------------
# Graph
#--------------------------------------------------

def Graph(name, *, profile:str|None=None, config:Config, dry_run:bool=False):
    client = Client(Resources(profile, config), rel.Compiler(config), name, dry_run=dry_run)
    client.install("pyrel_base", dsl.build.raw_task("""
        @inline
        def make_identity(x..., z) =
            hash128({x...}, x..., u) and
            hash_value_uint128_convert(u, z)
            from u

        @inline
        def pyrel_default[F, c](k..., v) =
            F(k..., v) or (not F(k..., _) and v = c)
    """))
    return dsl.Graph(client, name)



import relationalai as rai
from relationalai.std.graphs import Graph


# Test case taken from:
# https://github.com/RelationalAI/raicode/blob/8da410b949e8c47913a6faf4809234928fd8c238/test/RelLibraries/GraphAnalytics/graph_similarity_tests.jl#L238

model = rai.Model("cosineSimilarityTest", config=globals()["test_config"])
Node = model.Type("Node")

edge_list = [(11, 12), (12, 13), (13, 13), (12, 14), (14, 13)]
with model.rule(dynamic=True):
    for (x, y) in edge_list:
        node1 = Node.add(id=x)
        node2 = Node.add(id=y)
        node1.set(adjacent_to=node2)

graph = Graph(model, undirected=True)
graph.edges.extend(Node.adjacent_to)

with model.query() as select:
    node1 = Node(id=12)
    node2 = Node(id=14)
    similarity = graph.compute.cosine_similarity(node1, node2)
    response = select(node1.id, node2.id, similarity)


import relationalai as rai

model = rai.Model("FirstSnapshotTest", config=globals()["test_config"])

Person = model.Type("Person")

with model.rule():
    Person.add(name="Sam Watson")
    Person.add(name="Pete Vilter")
    Person.add(name="Josh Cole")

with model.query() as select:
    p = Person()
    select(p.name) # @NOTE: we don't actually need to use the result, it'll get snapshotted all the same



import os
import random
import pytest
import relationalai as rai
from relationalai.clients import config as cfg

@pytest.fixture(scope="session")
def engine_config():
    # If there's a local config file, use it, including
    # the engine specified there.
    config = cfg.Config()
    if config.file_path is not None:
        yield config
        return

    client_id = os.getenv('RAI_CLIENT_ID')
    client_secret = os.getenv('RAI_CLIENT_SECRET')

    if client_id is None or client_secret is None:
        raise ValueError("RAI_CLIENT_ID and RAI_CLIENT_SECRET must be set")

    # Otherwise, create a new engine and delete it afterwards.
    random_number = random.randint(1000000000, 9999999999)
    engine_name = f"pyrel-test-{random_number}"
    
    config = make_config(engine_name, client_id, client_secret)
    
    print(f"Creating engine {engine_name}")
    provider = rai.Resources(config=config)
    provider.create_engine(engine_name, 'XS')
    
    yield config
    
    print(f"Deleting engine {engine_name}")
    provider.delete_engine(engine_name)

def make_config(engine_name: str, client_id: str, client_secret: str) -> cfg.Config:
    # Running against prod
    return cfg.Config({
        'platform': "azure",
        'host': "azure.relationalai.com",
        'port': "443",
        'region': "us-east",
        'scheme': "https",
        'client_credentials_url': "https://login.relationalai.com/oauth/token",
        'client_id': client_id,
        'client_secret': client_secret,
        'engine': engine_name,
    })
